@@ -1,21 +1,69 @@
+#
+# Copyright 2016 University of Southern California
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#    http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+#
+import sys
+import hashlib
 import json
 from collections import OrderedDict
+
+# monkeypatch hashlib.algorithms_guaranteed on Python < 2.7.9, before importing bagit, since bagit-1.7.0 makes a
+# direct call to it at initialization time which will cause the module to fail to load - this can be removed if
+# bagit-python fixes the underlying issue
+try:
+    algorithms_guaranteed = hashlib.algorithms_guaranteed
+except AttributeError:  # pragma: no cover
+    # Python 2.7.0-2.7.8
+    hashlib.algorithms_guaranteed = set(hashlib.algorithms)
+
 import bagit
 from bagit import *
-from bagit import _, _can_read, _can_bag, _make_tagmanifest_file, _encode_filename, _decode_filename, \
-    _calc_hashes,_walk
-from bdbag import escape_url_path, VERSION, BAGIT_VERSION, PROJECT_URL
+from bagit import (_, _can_read, _can_bag, _make_tagmanifest_file, _encode_filename, _decode_filename, _calc_hashes,
+                   _walk)
+from bdbag import escape_uri, urlunquote, VERSION, BAGIT_VERSION, PROJECT_URL
 
 LOGGER = logging.getLogger(__name__)
 
+SUPPORTED_BAGIT_SPECS = ["0.97", "1.0"]
 
-def make_bag(bag_dir, bag_info=None, processes=1, checksums=None, encoding='utf-8', remote_entries=None):
+
+def parse_version(version):
+    try:
+        return tuple(int(i) for i in version.split(".", 1))
+    except ValueError:  # pragma: no cover
+        raise BagError(
+            _("Bag version numbers must be MAJOR.MINOR numbers, not %s") % version
+        )
+
+
+def make_bag(bag_dir,
+             bag_info=None,
+             processes=1,
+             checksums=None,
+             encoding='utf-8',
+             remote_entries=None,
+             spec_version="0.97"):
     """
     Convert a given directory into a bag. You can pass in arbitrary
     key/value pairs to put into the bag-info.txt metadata file as
     the bag_info dictionary.
     """
 
+    if spec_version not in SUPPORTED_BAGIT_SPECS:
+        raise RuntimeError(_("Unsupported BagIt specfication version: %s" % spec_version))
+    bag_version = parse_version(spec_version)
+
     if checksums is None:
         checksums = DEFAULT_CHECKSUMS
 
@@ -78,14 +126,15 @@ def make_bag(bag_dir, bag_info=None, processes=1, checksums=None, encoding='utf-
             # original directory
             os.chmod('data', os.stat(cwd).st_mode)
 
+            strict = True if bag_version >= (1, 0) else False
             validate_remote_entries(remote_entries, bag_dir)
             total_bytes, total_files = make_manifests(
-                'data', processes, algorithms=checksums, encoding=encoding, remote=remote_entries)
+                'data', processes, algorithms=checksums, encoding=encoding, remote=remote_entries, strict=strict)
 
             _make_fetch_file(bag_dir, remote_entries)
 
             LOGGER.info(_("Creating bagit.txt"))
-            txt = """BagIt-Version: 0.97\nTag-File-Character-Encoding: UTF-8\n"""
+            txt = """BagIt-Version: %s\nTag-File-Character-Encoding: UTF-8\n""" % spec_version
             with open_text_file('bagit.txt', 'w') as bagit_file:
                 bagit_file.write(txt)
 
@@ -113,7 +162,7 @@ def make_bag(bag_dir, bag_info=None, processes=1, checksums=None, encoding='utf-
     return BDBag(bag_dir)
 
 
-def make_manifests(data_dir, processes, algorithms=DEFAULT_CHECKSUMS, encoding='utf-8', remote=None):
+def make_manifests(data_dir, processes, algorithms=DEFAULT_CHECKSUMS, encoding='utf-8', remote=None, strict=False):
     LOGGER.info(_('Using %(process_count)d processes to generate manifests: %(algorithms)s'),
                 {'process_count': processes, 'algorithms': ', '.join(algorithms)})
 
@@ -155,6 +204,7 @@ def make_manifests(data_dir, processes, algorithms=DEFAULT_CHECKSUMS, encoding='
         for entry in batch:
             manifest_data.setdefault(entry[0], []).append(entry[1:])
 
+    file_entries = {}
     for algorithm, values in manifest_data.items():
         manifest_filename = 'manifest-%s.txt' % algorithm
 
@@ -163,6 +213,7 @@ def make_manifests(data_dir, processes, algorithms=DEFAULT_CHECKSUMS, encoding='
                 manifest.write("%s  %s\n" % (digest, _encode_filename(filename)))
                 num_files[algorithm] += 1
                 total_bytes[algorithm] += byte_count
+                file_entries[filename] = byte_count
 
     # We'll use sets of the values for the error checks and eventually return the payload oxum values:
     byte_value_set = set(total_bytes.values())
@@ -172,13 +223,21 @@ def make_manifests(data_dir, processes, algorithms=DEFAULT_CHECKSUMS, encoding='
     if not byte_value_set and not file_count_set:
         return 0, 0
 
-    if len(file_count_set) != 1:
-        raise RuntimeError(_('Expected the same number of files for each checksum'))
+    if strict:
+        if len(file_count_set) != 1:
+            raise RuntimeError(_('Expected the same number of files for each checksum'))
 
-    if len(byte_value_set) != 1:
-        raise RuntimeError(_('Expected the same number of bytes for each checksum'))
+        if len(byte_value_set) != 1:
+            raise RuntimeError(_('Expected the same number of bytes for each checksum'))
 
-    return byte_value_set.pop(), file_count_set.pop()
+        return byte_value_set.pop(), file_count_set.pop()
+
+    byte_total = file_total = 0
+    for file_size in file_entries.values():
+        file_total += 1
+        byte_total += file_size
+
+    return byte_total, file_total
 
 
 def validate_remote_entries(remote_entries, bag_path="."):
@@ -207,16 +266,6 @@ def make_remote_file_entry(remote_entries, filename, url, length, alg, digest):
     remote_entries[filename][alg] = digest
 
 
-def get_remote_algs(remote_entries):
-    algs = set()
-    if remote_entries:
-        for v in remote_entries.values():
-            for alg in CHECKSUM_ALGOS:
-                if alg in v:
-                    algs.add(alg)
-    return algs
-
-
 def _denormalize_filename(filename):
     if os.path.sep != '/':
         parts = filename.split(os.path.sep)
@@ -236,9 +285,9 @@ def _make_fetch_file(path, remote_entries):
     with open_text_file(fetch_file_path, 'w') as fetch_file:
         for filename in sorted(remote_entries.keys()):
             fetch_file.write("%s\t%s\t%s\n" %
-                             (escape_url_path(remote_entries[filename]['url'], "%/:=&?~#+!$,;'@()*[]"),
+                             (escape_uri(remote_entries[filename]['url']),
                               remote_entries[filename]['length'],
-                              _denormalize_filename(filename)))
+                              escape_uri(_denormalize_filename(filename))))
 
 
 def _find_tag_files(bag_dir):
@@ -252,7 +301,7 @@ def _find_tag_files(bag_dir):
                         continue
                     # remove everything up to the bag_dir directory
                     p = os.path.join(dir_name, filename)
-                    yield p
+                    yield _denormalize_filename(p)
 
 
 # monkeypatch bagit._find_tag_files with our version
@@ -291,7 +340,7 @@ class BagManifestConflict(BagError):
 
 class UnexpectedRemoteFile(ManifestErrorDetail):
     def __str__(self):
-        return _("%s exists in fetch.txt but is not in manifest") % self.path
+        return _("%s exists in fetch.txt but is not in manifest") % force_unicode(self.path)
 
 
 class BDBag(Bag):
@@ -302,6 +351,7 @@ class BDBag(Bag):
 
     def files_to_be_fetched(self, normalize=True):
         for f, size, path in self.fetch_entries():
+            path = urlunquote(path)
             if normalize:
                 yield os.path.normpath(path)
             else:
@@ -336,8 +386,11 @@ class BDBag(Bag):
             entry_path = os.path.normpath(filename.lstrip("*"))
             if entry_path in payload_entries:
                 for alg in self.algorithms:
-                    if payload_entries[entry_path].get(alg, None):
-                        self.add_remote_file(filename, url, length, alg, payload_entries[entry_path][alg])
+                    digest = payload_entries[entry_path].get(alg, None)
+                    remote_entry = self.remote_entries.get(filename)
+                    if remote_entry:
+                        continue
+                    self.add_remote_file(filename, url, length, alg, digest)
 
     def add_remote_file(self, filename, url, length, alg, digest):
         if alg not in self.algorithms:
@@ -388,12 +441,14 @@ class BDBag(Bag):
 
             # Generate new manifest files
             if manifests:
+                strict = True if self.version_info >= (1, 0) else False
                 self._sync_remote_entries_with_existing_fetch()
                 validate_remote_entries(self.remote_entries, self.path)
                 total_bytes, total_files = make_manifests('data', processes,
                                                           algorithms=self.algorithms,
                                                           encoding=self.encoding,
-                                                          remote=self.remote_entries)
+                                                          remote=self.remote_entries,
+                                                          strict=strict)
 
                 # Update fetch.txt
                 _make_fetch_file(self.path, self.remote_entries)
@@ -430,14 +485,28 @@ class BDBag(Bag):
         self._validate_structure()
         self._validate_bagittxt()
 
-        # this check is bypassed because it attempts to validate URL scheme and netloc,
-        # and bdbag supports non-standard URLs, e.g., "ark:/"
-        # self.validate_fetch()
+        self._validate_fetch()
 
         self._validate_contents(processes=processes, fast=fast, completeness_only=completeness_only, callback=callback)
 
         return True
 
+    def _validate_fetch(self):
+        """Validate the fetch.txt file
+
+        Raises `BagError` for errors and otherwise returns no value
+        """
+        for url, file_size, filename in self.fetch_entries():
+            # fetch_entries will raise a BagError for unsafe filenames
+            # so at this point we will check only that the URL is minimally
+            # well formed:
+            parsed_url = urlparse(url)
+
+            # only check for a scheme component since per the spec the URL field is actually a URI per
+            # RFC3986 (https://tools.ietf.org/html/rfc3986)
+            if not all(parsed_url.scheme):
+                raise BagError(_('Malformed URL in fetch.txt: %s') % url)
+
     def _validate_contents(self, processes=1, fast=False, completeness_only=False, callback=None):
         if fast and not self.has_oxum():
             raise BagValidationError(_('Fast validation requires bag-info.txt to include Payload-Oxum'))
@@ -507,7 +576,7 @@ class BDBag(Bag):
                         if not callback(count, totalHashes):
                             raise BaggingInterruptedError("Bag validation interrupted!")
 
-            else:
+            else:  # pragma: no cover
                 pool = None
                 try:
                     pool = multiprocessing.Pool(processes if processes else None, initializer=worker_init)
@@ -518,7 +587,7 @@ class BDBag(Bag):
         except BaggingInterruptedError:
             raise
         # Any unhandled exceptions are probably fatal
-        except:
+        except:  # pragma: no cover
             LOGGER.error(_("Unable to calculate file hashes for %s"), self)
             raise
 
